/* See LICENSE file for copyright and license details. */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <ftdi.h>
#include "util.h"

static struct ftdi_context ftdictx;
static struct ftdi_device_list *devlist;
static struct termios tio, savedtio;

static int bflag = 0;
static int lflag = 0;
static int dflag = 0;
static int vflag = 0;
static int pflag = 0;

static sig_atomic_t done = 0;

static void
interrupt(int sig)
{
	done = 1;
}

static int
ttyinit(void)
{
	long vdisable;

	errno = 0;
	vdisable = fpathconf(STDIN_FILENO, _PC_VDISABLE);
	if (vdisable < 0) {
		if (errno)
			eprintf("fpathconf: _PC_VDISABLE:");
		vdisable = 0;
	}

	tcgetattr(STDIN_FILENO, &tio);
	savedtio = tio;
	cfmakeraw(&tio);
	tio.c_cc[VMIN] = 0;
	tio.c_cc[VTIME] = 0;
	tio.c_lflag |= ISIG;
	tio.c_cc[VQUIT] = CTRL('\\');
	tio.c_cc[VINTR] = vdisable;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &tio);
	return 0;
}

static void
ttyrestore(void)
{
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &savedtio);
}

static int
loop(void)
{
	unsigned char buf[1];
	ssize_t n = 0;

	while (!done) {
		n = read(STDIN_FILENO, buf, sizeof(buf));
		if (n > 0)
			ftdi_write_data(&ftdictx, buf, n);
		n = ftdi_read_data(&ftdictx, buf, sizeof(buf));
		if (n < 0) {
			ttyrestore();
			enprintf(EXIT_FAILURE, "ftdi_read_data: read error: %s\n",
				 ftdi_get_error_string(&ftdictx));
		}
		write(STDOUT_FILENO, buf, n);
	}
	return n;
}

static void
listdevs(void)
{
	char manufacturer[256];
	char description[256];
	struct ftdi_device_list *iter;
	int i;
	int ret;

	for (i = 0, iter = devlist; iter; iter = iter->next, i++) {
		ret = ftdi_usb_get_strings(&ftdictx, iter->dev,
					   manufacturer, sizeof(manufacturer),
					   description, sizeof(description), NULL, 0);
		if (ret < 0)
			enprintf(EXIT_FAILURE, "ftdi_usb_get_strings: %s\n",
				 ftdi_get_error_string(&ftdictx));

		printf("device %d: manufacturer: %s, description: %s, vendor = %04x, product = %04x\n",
		       i, manufacturer, description, iter->dev->descriptor.idVendor,
		       iter->dev->descriptor.idProduct);
	}
}

static void
usage(void)
{
	fprintf(stderr, "usage: %s [-l] [-b baudrate] [-v vendor] [-p product] [-d device]\n", argv0);
	fprintf(stderr, "\t-l\tList devices\n");
	fprintf(stderr, "\t-b\tSet baudrate\n");
	fprintf(stderr, "\t-v\tSpecify vendor, defaults to 0x0403\n");
	fprintf(stderr, "\t-p\tSpecify product, defaults to 0x6001\n");
	fprintf(stderr, "\n\tNOTE: The break key is set to ^\\\n");
	exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
	int i;
	int baudrate = 115200;
	int dev = 0;
	int ret;
	struct sigaction sa;
	struct ftdi_device_list *iter;
	int vendor = 0x0403;
	int product = 0x6001;

	ARGBEGIN {
	case 'l':
		lflag = 1;
		break;
	case 'b':
		bflag = 1;
		baudrate = estrtol(EARGF(usage()), 10);
		break;
	case 'd':
		dflag = 1;
		dev = estrtol(EARGF(usage()), 10);
		break;
	case 'v':
		vflag = 1;
		vendor = estrtol(EARGF(usage()), 0);
		break;
	case 'p':
		pflag = 1;
		product = estrtol(EARGF(usage()), 0);
		break;
	default:
		usage();
	} ARGEND;

	if (!isatty(STDIN_FILENO))
		eprintf("isatty: stdin:");

	ret = ftdi_init(&ftdictx);
	if (ret < 0)
		enprintf(EXIT_FAILURE, "failed to initialize ftdi context\n");
	ret = ftdi_set_interface(&ftdictx, INTERFACE_ANY);
	if (ret < 0)
		enprintf(EXIT_FAILURE, "ftdi_set_interface: %s\n",
			 ftdi_get_error_string(&ftdictx));

	ret = ftdi_usb_find_all(&ftdictx, &devlist, vendor, product);
	if (ret < 0)
		enprintf(EXIT_FAILURE, "ftdi_usb_find_all: %s\n",
			 ftdi_get_error_string(&ftdictx));

	if (lflag) {
		listdevs();
		exit(EXIT_SUCCESS);
	}

	if (dflag) {
		for (i = 0, iter = devlist; iter; iter = iter->next, i++) {
			if (i == dev) {
				vendor = iter->dev->descriptor.idVendor;
				product = iter->dev->descriptor.idProduct;
				break;
			}
		}
		if (!iter)
			enprintf(EXIT_FAILURE, "Can't find device %d\n", dev);
	}

	ret = ftdi_usb_open(&ftdictx, vendor, product);
	if (ret < 0)
		enprintf(EXIT_FAILURE, "ftdi_usb_open: %04x:%04x: %s\n",
			 vendor, product, ftdi_get_error_string(&ftdictx));
	ret = ftdi_set_baudrate(&ftdictx, baudrate);
	if (ret < 0)
		enprintf(EXIT_FAILURE, "ftdi_set_baudrate: %d: %s\n",
			 baudrate, ftdi_get_error_string(&ftdictx));

	printf("Break key is set to ^\\\n");

	sa.sa_handler = interrupt;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sigaction(SIGQUIT, &sa, NULL);

	ttyinit();
	loop();
	ttyrestore();
	putchar('\n');
	ftdi_usb_close(&ftdictx);
	ftdi_list_free(&devlist);
	exit(EXIT_SUCCESS);
}
